#include <Arduino.h>
#include "HX711.h"
#include "soc/rtc.h"
#include "WiFi.h"
#include "Adafruit_MQTT.h"
#include "Adafruit_MQTT_Client.h"

// WiFi credentials
const char* ssid = "Eduard’s iPhone";
const char* password = "12345678";

// Adafruit IO credentials
#define AIO_SERVER      "io.adafruit.com"
#define AIO_SERVERPORT  1883                   // use 8883 for SSL
#define AIO_USERNAME    "eduardlev"
#define AIO_KEY         "aio_AvkN88yzFNWc7QkGp4Vvcnxbw6Cm"

// Motor A
int motor1Pin1 = 27;
int motor1Pin2 = 26;
int enable1Pin = 14;

// Setting PWM properties
const int freq = 30000;
const int pwmChannel = 0;
const int resolution = 8;
int dutyCycle = 80; // Maximum power for normal speed
int fineDutyCycle = 30; // Lower speed for fine adjustments

// HX711 circuit wiring
const int LOADCELL_DOUT_PIN = 16;
const int LOADCELL_SCK_PIN = 4;

HX711 scale;

const float targetWeightNormal = 18.0; // Target weight for normal in grams
const float targetWeightOverweight = 25.0; // Target weight for overweight in grams
const float stopBuffer = 1.0; // Buffer to switch to fine adjustment

WiFiClient client;
Adafruit_MQTT_Client mqtt(&client, AIO_SERVER, AIO_SERVERPORT, AIO_USERNAME, AIO_KEY);
Adafruit_MQTT_Subscribe feedMode = Adafruit_MQTT_Subscribe(&mqtt, AIO_USERNAME "/feeds/feedMode");

String feedModeValue = "none"; // Default value
bool feeding = false; // State to check if feeding is in progress

unsigned long lastFeedingTime = 0; // Timestamp for the last feeding
const unsigned long feedingInterval = 8 * 60 * 60 * 1000; // 8 hours interval

void setup() {
  // sets the pins as outputs:
  pinMode(motor1Pin1, OUTPUT);
  pinMode(motor1Pin2, OUTPUT);
  pinMode(enable1Pin, OUTPUT);

  // configure LED PWM functionalities
  ledcSetup(pwmChannel, freq, resolution);

  // attach the channel to the GPIO to be controlled
  ledcAttachPin(enable1Pin, pwmChannel);

  Serial.begin(115200);

  // Connect to WiFi
  Serial.print("Connecting to WiFi...");
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println("Connected to WiFi");

  // Initialize MQTT
  mqtt.subscribe(&feedMode);

  // Initialize HX711
  rtc_cpu_freq_config_t config;
  rtc_clk_cpu_freq_get_config(&config);
  rtc_clk_cpu_freq_to_config(RTC_CPU_FREQ_80M, &config);
  rtc_clk_cpu_freq_set_config_fast(&config);
  Serial.println("HX711 Demo");

  Serial.println("Initializing the scale");

  scale.begin(LOADCELL_DOUT_PIN, LOADCELL_SCK_PIN);

  Serial.println("Before setting up the scale:");
  Serial.print("read: \t\t");
  Serial.println(scale.read()); // print a raw reading from the ADC

  Serial.print("read average: \t\t");
  Serial.println(scale.read_average(20)); // print the average of 20 readings from the ADC

  Serial.print("get value: \t\t");
  Serial.println(scale.get_value(5)); // print the average of 5 readings from the ADC minus the tare weight (not set yet)

  Serial.print("get units: \t\t");
  Serial.println(scale.get_units(5), 1); // print the average of 5 readings from the ADC minus tare weight (not set) divided
  // by the SCALE parameter (not set yet)

  scale.set_scale(1177.713); // this value is obtained by calibrating the scale with known weights; see the README for details
  scale.tare(); // reset the scale to 0

  Serial.println("After setting up the scale:");

  Serial.print("read: \t\t");
  Serial.println(scale.read()); // print a raw reading from the ADC

  Serial.print("read average: \t\t");
  Serial.println(scale.read_average(20)); // print the average of 20 readings from the ADC

  Serial.print("get value: \t\t");
  Serial.println(scale.get_value(5)); // print the average of 5 readings from the ADC minus the tare weight, set with tare()

  Serial.print("get units: \t\t");
  Serial.println(scale.get_units(5), 1); // print the average of 5 readings from the ADC minus tare weight, divided
  // by the SCALE parameter set with set_scale

  Serial.println("Readings:");
}

void loop() {
  // Ensure the connection to Adafruit IO is working
  MQTT_connect();

  // Check for feed updates
  Adafruit_MQTT_Subscribe *subscription;
  while ((subscription = mqtt.readSubscription(5000))) {
    if (subscription == &feedMode) {
      feedModeValue = (char *)feedMode.lastread;
      Serial.print("Feed Mode: ");
      Serial.println(feedModeValue);
      feeding = true; // Start feeding process
      lastFeedingTime = 0; // Reset last feeding time to allow immediate feeding
    }
  }

  if (feeding) {
    unsigned long currentTime = millis();
    Serial.print("Current Time: ");
    Serial.println(currentTime);
    Serial.print("Last Feeding Time: ");
    Serial.println(lastFeedingTime);

    if (currentTime - lastFeedingTime >= feedingInterval || lastFeedingTime == 0) {
      float targetWeight = (feedModeValue == "normal") ? targetWeightNormal : targetWeightOverweight;
      scale.power_up();

      float weight = 0;
      unsigned long startTime = millis(); // Track the start time

      while (true) {
        weight = scale.get_units(10);

        Serial.print("Current weight: ");
        Serial.println(weight, 1);

        if (weight < (targetWeight - stopBuffer)) {
          // Move the DC motor forward at normal speed
          Serial.println("Moving Forward at normal speed");
          digitalWrite(motor1Pin1, HIGH);
          digitalWrite(motor1Pin2, LOW);
          ledcWrite(pwmChannel, dutyCycle); // Set speed
        } else if (weight < targetWeight) {
          // Move the DC motor forward at fine speed
          Serial.println("Moving Forward at fine speed");
          digitalWrite(motor1Pin1, HIGH);
          digitalWrite(motor1Pin2, LOW);
          ledcWrite(pwmChannel, fineDutyCycle); // Set fine speed
        } else {
          // Stop the DC motor
          Serial.println("Motor stopped");
          digitalWrite(motor1Pin1, LOW);
          digitalWrite(motor1Pin2, LOW);
          ledcWrite(pwmChannel, 0); // Stop PWM

          delay(100); // Short delay to allow food to fall completely

          // Double-check weight to ensure accuracy
          weight = scale.get_units(10);
          if (weight >= targetWeight) {
            Serial.println("Target weight reached");
            scale.power_down(); // Power down the scale only after stopping the motor
            lastFeedingTime = currentTime; // Update last feeding time
            feeding = false; // Reset feeding state
            feedModeValue = "none"; // Clear feed mode
            break; // Exit the loop
          }
        }

        // Check for timeout to avoid infinite loop
        if (millis() - startTime > 30000) { // 30 seconds timeout
          Serial.println("Timeout reached, stopping motor");
          digitalWrite(motor1Pin1, LOW);
          digitalWrite(motor1Pin2, LOW);
          ledcWrite(pwmChannel, 0); // Stop PWM
          feeding = false; // Reset feeding state
          feedModeValue = "none"; // Clear feed mode
          break; // Exit the loop
        }

        delay(100); // Delay between weight measurements
      }
    }
  }
}

void MQTT_connect() {
  int8_t ret;

  // Stop if already connected.
  if (mqtt.connected()) {
    return;
  }

  Serial.print("Connecting to Adafruit IO... ");
  while ((ret = mqtt.connect()) != 0) { // connect will return 0 for connected
    Serial.println(mqtt.connectErrorString(ret));
    Serial.println("Retrying connection in 5 seconds...");
    mqtt.disconnect();
    delay(5000);  // wait 5 seconds
  }
  Serial.println("Connected to Adafruit IO!");
}
